/*
 *  libxcpc by Davide Libenzi (Exception and resource handling in C)
 *  Copyright (C) 2002  Davide Libenzi
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#if !defined(_LIBXCPC_H)
#define _LIBXCPC_H

#include <stdio.h>
#include <setjmp.h>


#if defined(__cplusplus)
#define XCEXTC extern "C"
#else
#define XCEXTC
#endif

#define XCAPI XCEXTC

#define XCHTYPE(n) typedef struct s_##n { const int x; } *n

#define XCPC_EXCEPT_DATA __tctx.data
#define XCPC_EXCEPTION __exno

#define XCPC_THROW(c, e, d) xcpc_do_throw(c, e, d)

#define XCPC_RETHROW(c) xcpc_do_rethrow(c)

#define XCPC_TRY(c) do { \
	xcpc_tryctx __tctx; \
	volatile int __exno, __caught; \
	do { \
		xcpc_push_tryctx(c, &__tctx); \
		__exno = setjmp(__tctx.context); \
		__caught = 0; \
		if (__exno == 0) {

#define XCPC_CATCH(e) } else if ((e) == __exno) { \
	__caught++;

#define XCPC_CATCH_ANY } else { \
	__caught++;

#define XCPC_EPILOGUE do { \
	if (__exno != 0 && !__caught) \
		xcpc_do_rethrow(__tctx.ctx); \
	else \
		xcpc_pop_tryctx(__tctx.ctx); \
} while (0)

#define XCPC_END_TRY } \
	XCPC_EPILOGUE; \
	} while (0); \
	} while (0)

#define XCPC_RETURNV(r) do { \
	XCPC_EPILOGUE; \
	return r; \
} while (0)

#define XCPC_RETURN do { \
	XCPC_EPILOGUE; \
	return; \
} while (0)



XCHTYPE(xcpc_ctx);
XCHTYPE(xcpc_res);

enum xcpc_exceptions {
	/* Out of memory */
	XCPCE_MEMORY = 1,

		/* Trying to pop a context with empty list */
		XCPCE_NOEXCTX,

		/* Try to pop wrong contaxt */
		XCPCE_POPWRONGCTX,

		/* Trying to re-parent root */
		XCPCE_REPARENT_ROOT,

		/* Trying to re-parent unrealted contexts */
		XCPCE_REPARENT_UNREL,

		/*
		 * Failed fopen(3): DATA = path
		 */
		XCPCE_FOPEN,

		/*
		 * Failed fdopen(3): DATA = fd
		 */
		XCPCE_FDOPEN,

		/*
		 * Failed freopen(3): DATA = path
		 */
		XCPCE_FREOPEN,

		/*
		 * Failed open(2): DATA = path
		 */
		XCPCE_OPEN,

		/*
		 * Failed mmap(2)
		 */
		XCPCE_MMAP,

		/*
		 * Failed write(2): DATA = fd
		 */
		XCPCE_WRITE,

		/*
		 * Failed read(2): DATA = fd
		 */
		XCPCE_READ,

		/*
		 * Failed fwrite(3): DATA = stream
		 */
		XCPCE_FWRITE,

		/*
		 * Failed fread(3): DATA = stream
		 */
		XCPCE_FREAD,

		/*
		 * Failed stat(2): DATA = path
		 */
		XCPCE_STAT,

		/*
		 * Failed fstat(2): DATA = fd
		 */
		XCPCE_FSTAT,

		/*
		 * Failed mkdir(2): DATA = path
		 */
		XCPCE_MKDIR,

		/*
		 * Failed rmdir(2): DATA = path
		 */
		XCPCE_RMDIR,

		/*
		 * Failed remove(3): DATA = path
		 */
		XCPCE_REMOVE,

		/*
		 * Failed opendir(3): DATA = path
		 */
		XCPCE_OPENDIR,

		/*
		 * Failed pipe(2)
		 */
		XCPCE_PIPE,

		/*
		 * Failed socket(2)
		 */
		XCPCE_SOCKET,

		/*
		 * Failed bind(2): DATA = struct sockaddr
		 */
		XCPCE_BIND,

		/*
		 * Failed connect(2): DATA = struct sockaddr
		 */
		XCPCE_CONNECT,

		/*
		 * Failed truncate(2): DATA = path
		 */
		XCPCE_TRUNCATE,

		/*
		 * Failed ftruncate(2): DATA = fd
		 */
		XCPCE_FTRUNCATE,

		/*
		 * Failed lseek(2): DATA = fd
		 */
		XCPCE_LSEEK,

		XCPCE_USER
};

typedef struct s_xcpc_listhead {
	struct s_xcpc_listhead *next;
	struct s_xcpc_listhead *prev;
} xcpc_listhead;

typedef struct s_xcpc_tryctx {
	xcpc_listhead lnk;
	struct s_xcpc_tryctx *parent;
	volatile xcpc_ctx ctx;
	jmp_buf context;
	volatile int exno;
	volatile void *data;
	xcpc_listhead ctxlst;
	xcpc_listhead reslst;
} xcpc_tryctx;



/*
 * Core libxcpc functions ...
 */
XCAPI void xcpc_set_exitproc(void (*proc)(int));
XCAPI void xcpc_push_tryctx(xcpc_ctx ctx, xcpc_tryctx *tctx);
XCAPI xcpc_tryctx *xcpc_pop_tryctx(xcpc_ctx ctx);
XCAPI void xcpc_do_throw(xcpc_ctx ctx, int exno, void *data);
XCAPI void xcpc_do_rethrow(xcpc_ctx ctx);
XCAPI xcpc_ctx xcpc_context_create(xcpc_ctx pctx);
XCAPI void xcpc_context_reparent(xcpc_ctx ctx, xcpc_ctx pctx);
XCAPI void xcpc_context_free(xcpc_ctx ctx);
XCAPI xcpc_ctx xcpc_context_parent(xcpc_ctx ctx);
XCAPI xcpc_ctx xcpc_context_root(xcpc_ctx ctx);
XCAPI xcpc_ctx xcpc_context_exhandler(xcpc_ctx ctx);
XCAPI xcpc_res xcpc_resource_add(xcpc_ctx ctx, void *data, void (*free)(void *));
XCAPI void xcpc_resource_movehead(xcpc_res res);
XCAPI void xcpc_resource_movetail(xcpc_res res);
XCAPI void xcpc_resource_moveafter(xcpc_res res, xcpc_res rres);
XCAPI void xcpc_resource_movebefore(xcpc_res res, xcpc_res rres);
XCAPI xcpc_ctx xcpc_resource_ctx(xcpc_res res);
XCAPI void *xcpc_resource_data(xcpc_res res);
XCAPI void xcpc_resource_del(xcpc_res res, int do_free);
XCAPI void xcpc_resource_reparent(xcpc_res res, xcpc_ctx pctx);
XCAPI void xcpc_resource_set(xcpc_res res, int do_free, void *data,
			     void (*free)(void *));

/*
 * Wrapper functions for  libxcpc ...
 */
struct stat;
struct FILE;
struct sockaddr;

XCAPI void *xcpc__malloc(xcpc_ctx ctx, xcpc_res *pres, long size);
XCAPI void *xcpc__realloc(xcpc_res res, long size);
XCAPI char *xcpc__strdup(xcpc_ctx ctx, xcpc_res *pres, char const *str);
XCAPI FILE *xcpc__fopen(xcpc_ctx ctx, xcpc_res *pres, char const *path,
			char const *mode);
XCAPI FILE *xcpc__fdopen(xcpc_ctx ctx, xcpc_res *pres, int fd,
			 char const *mode);
XCAPI FILE *xcpc__freopen(xcpc_res res, char const *path,
			  char const *mode);
XCAPI int xcpc__open(xcpc_ctx ctx, xcpc_res *pres, char const *path,
		     int flags, int mode);
XCAPI void xcpc__remove(xcpc_ctx ctx, const char *path);
XCAPI void xcpc__push_remove(xcpc_ctx ctx, xcpc_res *pres, char const *path);
XCAPI void *xcpc__mmap(xcpc_ctx ctx, xcpc_res *pres, void *start, size_t length,
		       int prot, int flags, int fd, off_t offset);
XCAPI ssize_t xcpc__write(xcpc_ctx ctx, int fd, const void *buf, size_t count);
XCAPI ssize_t xcpc__read(xcpc_ctx ctx, int fd, void *buf, size_t count);
XCAPI size_t xcpc__fwrite(xcpc_ctx ctx, const void *ptr, size_t size, size_t nmemb,
			  FILE *stream);
XCAPI size_t xcpc__fread(xcpc_ctx ctx, void *ptr, size_t size, size_t nmemb,
			 FILE *stream);
XCAPI void xcpc__stat(xcpc_ctx ctx, const char *path, struct stat *buf);
XCAPI void xcpc__fstat(xcpc_ctx ctx, int fd, struct stat *buf);
XCAPI void xcpc__mkdir(xcpc_ctx ctx, const char *path, mode_t mode);
XCAPI void xcpc__rmdir(xcpc_ctx ctx, const char *path);
XCAPI void *xcpc__opendir(xcpc_ctx ctx, const char *path);
XCAPI void xcpc__pipe(xcpc_ctx ctx, int *fds);
XCAPI int xcpc__socket(xcpc_ctx ctx, int domain, int type, int protocol);
XCAPI void xcpc__bind(xcpc_ctx ctx, int sfd, const struct sockaddr *addr,
		      int alen);
XCAPI void xcpc__connect(xcpc_ctx ctx, int sfd, const struct sockaddr *addr,
			 int alen);
XCAPI void xcpc__truncate(xcpc_ctx ctx, char const *path, off_t len);
XCAPI void xcpc__ftruncate(xcpc_ctx ctx, int fd, off_t len);
XCAPI off_t xcpc__lseek(xcpc_ctx ctx, int fd, off_t off, int whence);

#endif

