/*
 *  libxcpc by Davide Libenzi (Exception and resource handling in C)
 *  Copyright (C) 2002  Davide Libenzi
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Davide Libenzi <davidel@xmailserver.org>
 *
 */

#include "xcpc-include.h"



typedef struct s_xcpc_context {
	struct s_xcpc_context *parent;
	struct s_xcpc_context *root;
	xcpc_listhead lnk;
	xcpc_listhead ctxlst;
	xcpc_listhead reslst;
	xcpc_listhead exlst;
	xcpc_tryctx *tctx;
} xcpc_context;

typedef struct s_xcpc_resource {
	xcpc_listhead lnk;
	xcpc_context *xctx;
	void *data;
	void (*free)(void *);
} xcpc_resource;



static xcpc_tryctx *xcpc_get_handler(xcpc_context *xctx, int exno);
static void xcpc_free_resource(xcpc_resource *xres, int do_free);
static void xcpc_free_ctx(xcpc_context *xctx);
static void xcpc_free_reslist(xcpc_listhead *head);
static void xcpc_free_ctxlist(xcpc_listhead *head);
static xcpc_listhead *xcpc_reslst_head(xcpc_context *xctx);
static xcpc_listhead *xcpc_ctxlst_head(xcpc_context *xctx);
static xcpc_resource *xcpc_alloc_res(xcpc_context *xctx, void *data,
				     void (*free)(void *));
static xcpc_context *xcpc_alloc_ctx(xcpc_context *xctx);
static void xcpc_panic(int exno);



#ifdef HAVE_EXIT
static void (*xcpc__exit)(int) = (void (*)(int)) exit;
#else
static void (*xcpc__exit)(int);
#endif



XCAPI void xcpc_set_exitproc(void (*proc)(int)) {

	xcpc__exit = proc;
}

static xcpc_tryctx *xcpc_get_handler(xcpc_context *xctx, int exno) {
	xcpc_listhead *pos;

	if ((pos = XCM_LIST_LAST(&xctx->root->exlst)) == NULL)
		xcpc_panic(exno);

	return XCM_LIST_ENTRY(pos, xcpc_tryctx, lnk);
}

static void xcpc_free_resource(xcpc_resource *xres, int do_free) {

	XCM_LIST_DEL(&xres->lnk);
	if (do_free && xres->free != NULL)
		(*xres->free)(xres->data);
	free(xres);
}

static void xcpc_free_ctx(xcpc_context *xctx) {

	XCM_LIST_DEL(&xctx->lnk);
	xcpc_free_ctxlist(&xctx->ctxlst);
	xcpc_free_reslist(&xctx->reslst);
	free(xctx);
}

static void xcpc_free_reslist(xcpc_listhead *head) {
	xcpc_listhead *pos;
	xcpc_resource *xres;

	while ((pos = XCM_LIST_FIRST(head)) != NULL) {
		xres = XCM_LIST_ENTRY(pos, xcpc_resource, lnk);
		xcpc_free_resource(xres, 1);
	}
}

static void xcpc_free_ctxlist(xcpc_listhead *head) {
	xcpc_listhead *pos;
	xcpc_context *xctx;

	while ((pos = XCM_LIST_FIRST(head)) != NULL) {
		xctx = XCM_LIST_ENTRY(pos, xcpc_context, lnk);
		xcpc_free_ctx(xctx);
	}
}

static xcpc_listhead *xcpc_reslst_head(xcpc_context *xctx) {

	return xctx->tctx == NULL ? &xctx->reslst: &xctx->tctx->reslst;
}

static xcpc_listhead *xcpc_ctxlst_head(xcpc_context *xctx) {

	return xctx->tctx == NULL ? &xctx->ctxlst: &xctx->tctx->ctxlst;
}

static xcpc_resource *xcpc_alloc_res(xcpc_context *xctx, void *data,
				     void (*free)(void *)) {
	xcpc_resource *xres;
	xcpc_listhead *head;

	if ((xres = (xcpc_resource *) malloc(sizeof(xcpc_resource))) == NULL) {
		if (free != NULL)
			(*free)(data);
		xcpc_do_throw((xcpc_ctx) xctx, XCPCE_MEMORY, NULL);
	}
	xres->xctx = xctx;
	xres->data = data;
	xres->free = free;
	head = xcpc_reslst_head(xctx);
	XCM_LIST_ADDH(&xres->lnk, head);

	return xres;
}

static xcpc_context *xcpc_alloc_ctx(xcpc_context *xctx) {
	xcpc_context *nctx;
	xcpc_listhead *head;

	if ((nctx = (xcpc_context *) malloc(sizeof(xcpc_context))) == NULL)
		xcpc_do_throw((xcpc_ctx) xctx, XCPCE_MEMORY, NULL);
	XCM_INIT_LIST_HEAD(&nctx->lnk);
	XCM_INIT_LIST_HEAD(&nctx->ctxlst);
	XCM_INIT_LIST_HEAD(&nctx->reslst);
	XCM_INIT_LIST_HEAD(&nctx->exlst);
	nctx->parent = xctx;
	nctx->tctx = NULL;
	if (xctx != NULL) {
		nctx->root = xctx->root;
		head = xcpc_ctxlst_head(xctx);
		XCM_LIST_ADDH(&nctx->lnk, head);
	} else
		nctx->root = nctx;

	return nctx;
}

XCAPI void xcpc_push_tryctx(xcpc_ctx ctx, xcpc_tryctx *tctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_context *rctx = xctx->root;

	tctx->parent = xctx->tctx;
	xctx->tctx = tctx;

	tctx->ctx = ctx;
	tctx->exno = 0;
	tctx->data = NULL;
	XCM_INIT_LIST_HEAD(&tctx->ctxlst);
	XCM_INIT_LIST_HEAD(&tctx->reslst);
	XCM_LIST_ADDT(&tctx->lnk, &rctx->exlst);
}

XCAPI xcpc_tryctx *xcpc_pop_tryctx(xcpc_ctx ctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_context *rctx = xctx->root;
	xcpc_tryctx *tctx;

	tctx = xcpc_get_handler(xctx, XCPCE_NOEXCTX);
	if (tctx->ctx != ctx)
		xcpc_panic(XCPCE_POPWRONGCTX);
	XCM_LIST_DEL(&tctx->lnk);
	xctx->tctx = tctx->parent;
	if (tctx->exno) {
		xcpc_free_ctxlist(&tctx->ctxlst);
		xcpc_free_reslist(&tctx->reslst);
	} else {
		XCM_LIST_SPLICE(&tctx->ctxlst, &xctx->ctxlst);
		XCM_LIST_SPLICE(&tctx->reslst, &xctx->reslst);
	}

	return tctx;
}

static void xcpc_panic(int exno) {

	if (xcpc__exit != NULL)
		(*xcpc__exit)(exno);

	/* If no exit proc is set, this is crash time! */
	*(int *) 0 = 0;
}

XCAPI void xcpc_do_throw(xcpc_ctx ctx, int exno, void *data) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_tryctx *tctx;

	tctx = xcpc_get_handler(xctx, exno);
	if (tctx->exno != 0) {
		/*
		 * Another exception while handling a previous one.
		 */

	}
	tctx->exno = exno;
	tctx->data = data;

	longjmp(tctx->context, exno);
}

XCAPI void xcpc_do_rethrow(xcpc_ctx ctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_tryctx *otctx, *tctx;

	otctx = xcpc_get_handler(xctx, XCPCE_NOEXCTX);
	if (otctx->ctx != ctx)
		xcpc_panic(XCPCE_POPWRONGCTX);
	XCM_LIST_DEL(&otctx->lnk);
	xctx->tctx = otctx->parent;

	tctx = xcpc_get_handler(xctx, otctx->exno);

	XCM_LIST_SPLICE(&otctx->ctxlst, &tctx->ctxlst);
	XCM_LIST_SPLICE(&otctx->reslst, &tctx->reslst);

	tctx->exno = otctx->exno;
	tctx->data = otctx->data;

	longjmp(tctx->context, tctx->exno);
}

XCAPI xcpc_ctx xcpc_context_create(xcpc_ctx pctx) {
	xcpc_context *xpctx = (xcpc_context *) pctx;
	xcpc_context *xctx;

	xctx = xcpc_alloc_ctx(xpctx);

	return (xcpc_ctx) xctx;
}

XCAPI void xcpc_context_reparent(xcpc_ctx ctx, xcpc_ctx pctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_context *xpctx = (xcpc_context *) pctx;
	xcpc_listhead *head;

	/*
	 * Root context cannot be re-parented!
	 */
	if (xctx->root == xctx)
		xcpc_do_throw(ctx, XCPCE_REPARENT_ROOT, NULL);
	if (xctx->root != xpctx->root)
		xcpc_do_throw(ctx, XCPCE_REPARENT_UNREL, NULL);
	XCM_LIST_DEL(&xctx->lnk);
	xctx->parent = xpctx;
	head = xcpc_ctxlst_head(xpctx);
	XCM_LIST_ADDH(&xctx->lnk, head);
}

XCAPI void xcpc_context_free(xcpc_ctx ctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;

	xcpc_free_ctx(xctx);
}

XCAPI xcpc_ctx xcpc_context_parent(xcpc_ctx ctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;

	return (xcpc_ctx) xctx->parent;
}

XCAPI xcpc_ctx xcpc_context_root(xcpc_ctx ctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;

	return (xcpc_ctx) xctx->root;
}

XCAPI xcpc_ctx xcpc_context_exhandler(xcpc_ctx ctx) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_tryctx *tctx;

	tctx = xcpc_get_handler(xctx, XCPCE_NOEXCTX);

	return tctx->ctx;
}

XCAPI xcpc_res xcpc_resource_add(xcpc_ctx ctx, void *data, void (*free)(void *)) {
	xcpc_context *xctx = (xcpc_context *) ctx;
	xcpc_resource *xres;

	xres = xcpc_alloc_res(xctx, data, free);

	return (xcpc_res) xres;
}

XCAPI void xcpc_resource_movehead(xcpc_res res) {
	xcpc_resource *xres = (xcpc_resource *) res;
	xcpc_listhead *head;

	XCM_LIST_DEL(&xres->lnk);
	head = xcpc_reslst_head(xres->xctx);
	XCM_LIST_ADDH(&xres->lnk, head);
}

XCAPI void xcpc_resource_movetail(xcpc_res res) {
	xcpc_resource *xres = (xcpc_resource *) res;
	xcpc_listhead *head;

	XCM_LIST_DEL(&xres->lnk);
	head = xcpc_reslst_head(xres->xctx);
	XCM_LIST_ADDT(&xres->lnk, head);
}

XCAPI void xcpc_resource_moveafter(xcpc_res res, xcpc_res rres) {
	xcpc_resource *xres = (xcpc_resource *) res;
	xcpc_resource *xrres = (xcpc_resource *) rres;

	XCM_LIST_DEL(&xres->lnk);
	XCM_LIST_ADD(&xres->lnk, &xrres->lnk, xrres->lnk.next);
	xres->xctx = xrres->xctx;
}

XCAPI void xcpc_resource_movebefore(xcpc_res res, xcpc_res rres) {
	xcpc_resource *xres = (xcpc_resource *) res;
	xcpc_resource *xrres = (xcpc_resource *) rres;

	XCM_LIST_DEL(&xres->lnk);
	XCM_LIST_ADD(&xres->lnk, xrres->lnk.prev, &xrres->lnk);
	xres->xctx = xrres->xctx;
}

XCAPI xcpc_ctx xcpc_resource_ctx(xcpc_res res) {
	xcpc_resource *xres = (xcpc_resource *) res;

	return (xcpc_ctx) xres->xctx;
}

XCAPI void *xcpc_resource_data(xcpc_res res) {
	xcpc_resource *xres = (xcpc_resource *) res;

	return xres->data;
}

XCAPI void xcpc_resource_del(xcpc_res res, int do_free) {
	xcpc_resource *xres = (xcpc_resource *) res;

	xcpc_free_resource(xres, do_free);
}

XCAPI void xcpc_resource_reparent(xcpc_res res, xcpc_ctx pctx) {
	xcpc_resource *xres = (xcpc_resource *) res;
	xcpc_context *xpctx = (xcpc_context *) pctx;
	xcpc_listhead *head;

	XCM_LIST_DEL(&xres->lnk);
	head = xcpc_reslst_head(xpctx);
	XCM_LIST_ADDH(&xres->lnk, head);
	xres->xctx = xpctx;
}

XCAPI void xcpc_resource_set(xcpc_res res, int do_free, void *data,
			     void (*free)(void *)) {
	xcpc_resource *xres = (xcpc_resource *) res;

	if (do_free && xres->free != NULL)
		(*xres->free)(xres->data);
	xres->data = data;
	xres->free = free;
}

